﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;

namespace WindowsFormsApplication1
{
    public partial class Form1 : Form
    {
        private LevelData levels;
        int currentLevel;
        bool modifiedSinceLastSave;
        int selectedBrick;
        string fileName;
        Bitmap[] brickBitmaps;

        public Form1()
        {
            InitializeComponent();
        }

        private void drawBrick(Graphics g, int x, int y)
        {
            if (levels.levelType == LevelType.TI84CE)
            {
                int brick = levels.levels[currentLevel][x, y];
                drawColorBrick(x * 32, y * 16, g, brick);
            }
            else
            {
                Rectangle rect = new Rectangle(x * 32, y * 16, 32, 16);
                int blankSides = (20 - levels.GetWidth()) / 2;
                if (x < blankSides || x + blankSides > 19)
                {
                    g.FillRectangle(new SolidBrush(Color.LightGray), rect);

                }
                else if (y >= levels.GetHeight())
                {
                    g.FillRectangle(new SolidBrush(Color.White), rect);
                }
                else
                {
                    int adjusted = x - blankSides;
                    int brick = levels.levels[currentLevel][adjusted, y];
                    if (levels.levelType != LevelType.TI89) brick += 16;
                    g.DrawImageUnscaled(brickBitmaps[brick], rect.Left, rect.Top);
                }
            }
        }

        private void panel1_Paint(object sender, PaintEventArgs e)
        {
            for (int x = 0; x < 20; x++)
            {
                for (int y = 0; y < 16; y++)
                {
                    drawBrick(e.Graphics, x, y);
                }
            }
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            brickBitmaps = new Bitmap[48];
            for (int b = 0; b < 48; b++)
                brickBitmaps[b] = WindowsFormsApplication1.Properties.Resources.bricks.Clone(new Rectangle(b * 33, 0, 32, 16), System.Drawing.Imaging.PixelFormat.Format32bppPArgb);
            clearData(LevelType.TI84CE);
        }

        private void clearData(LevelType levelType)
        {
            if (modifiedSinceLastSave && levels.IsNonTrivial() && !ShouldContinue()) return;
            levels = new LevelData(levelType);
            selectedBrick = levelType == LevelType.TI89 ? 4 : 1;
            currentLevel = 0;
            refreshCount();
        }

        private bool ShouldContinue()
        {
            return MessageBox.Show("Current level data is not saved.  Do you want to continue, erasing it?", "Monster / Orzunoid Editor", MessageBoxButtons.YesNo) == DialogResult.Yes;
        }

        private bool ConfirmClear()
        {
            return MessageBox.Show("Are you sure you want to erase this level?", "Monster / Orzunoid Editor", MessageBoxButtons.YesNo) == DialogResult.Yes;
        }

        private void refreshCount()
        {
            String count = "Now editing level " + (currentLevel + 1) + " of " + levels.levels.Count;
            labelCount.Text = count;
            buttonDelete.Enabled = levels.levels.Count > levels.GetLevelMin();
            buttonInsert.Enabled = levels.levels.Count < levels.GetLevelMax();
            buttonDuplicate.Enabled = levels.levels.Count < levels.GetLevelMax();
            labelType.Text = levels.Description();
            Refresh();
            panelColor.Visible = levels.levelType == LevelType.TI84CE;
        }

        private void orzunoidLevelsTI82TI83ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            clearData(LevelType.TI82);
        }

        private void monsterLevelsTI89TI92PlusVoyage200ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            clearData(LevelType.TI89);
        }

        private void orzunoidLevelsTI85TI86ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            clearData(LevelType.TI85);
        }

        private void button1_Click(object sender, EventArgs e)
        {
            if (currentLevel + 1 == levels.levels.Count)
            {
                if (!levels.AddLevel())
                {
                    return;
                }
            }
            currentLevel++;
            refreshCount();
        }

        private Color getColor(int color, int brightness)
        {
            if (brightness == 0) return Color.Black;
            Color c = new Color[] {
                          Color.FromArgb(192, 192, 192),
                          Color.FromArgb(130, 130, 130),
                          Color.Red,
                          Color.FromArgb(253, 40, 0),
                          Color.FromArgb(255, 80, 0),
                          Color.FromArgb(255, 120, 0),
                          Color.FromArgb(255, 160, 0),
                          Color.FromArgb(255, 200, 0),
                          Color.FromArgb(246, 246, 1),
                          Color.FromArgb(0, 240, 0),
                          Color.Green,
                          Color.Cyan,
                          Color.FromArgb(0, 35, 255),
                          Color.BlueViolet,
                          Color.Violet,
                          Color.FromArgb(255, 255, 0, 90)
                      }[color];
            if (brightness == 1) return Color.FromArgb(c.R * 3 / 4, c.G * 3 / 4, c.B * 3 / 4);
            if (brightness == 2) return c;
            if (brightness == 3)
            {
                if (color == 0) return Color.FromArgb(255, 255, 255);
                if (color == 12) return Color.FromArgb(0, 127, 255);
                if (c.B == 0) return Color.FromArgb((c.R + 255) / 2, (c.G + 255) / 2, 0);
                return Color.FromArgb((c.R * 2 + 255) / 3, (c.G * 2 + 255) / 3, (c.B * 2 + 255) / 3);
            }
            return Color.Black;
        }

        private int getPixel(int brick, int x, int y)
        {
            bool topOrLeft = (x == 0 && y != 7) || (y == 0 && x != 15);
            bool almostTopOrLeft = (x == 1 && y < 6) || (y == 1 && x < 14);
            bool bottomOrRight = (x == 15 && y != 0) || (y == 7 && x != 0);
            bool almostBottomOrRight = (x == 14 && y > 1) || (y == 6 && x > 1);
            if (brick == 1) // standard brick
            {
                if (topOrLeft) return 3;
                if (bottomOrRight) return 1;
                if (almostTopOrLeft) return 1;
                if (almostBottomOrRight) return 3;
                return 2;
            }
            if (brick == 2) // double-hit brick
            {
                if (topOrLeft) return 3;
                if (bottomOrRight) return 1;
                return 2;
            }
            if (brick == 3) // unbreakable brick
            {
                if (y < 4 && x < 15 - y) return 2;
                if (y >= 4 && x < 7 - y) return 2;
                return 1;
            }
            if (brick == 4) // triple-hit brick
            {
                if (topOrLeft || almostTopOrLeft) return 3;
                if (bottomOrRight || almostBottomOrRight) return 1;
                return 2;
            }
            if (brick == 5) return pixelFromResource(11, x, y); // double unbreakable top
            if (brick == 6) return pixelFromResource(12, x, y); // double unbreakable bottom
            if (brick == 7) return pixelFromResource(x > 7 ? 8 : 7, x, y); // top of tall brick
            if (brick == 8) return pixelFromResource(x > 7 ? 9 : 10, x, y); // bottom of tall brick
            if (brick >= 9 && brick <= 12) return pixelFromResource(brick - 2, x, y); // quad-size brick
            if (brick == 13) return pixelFromResource(x > 7 ? 8 : 7, x, 4); // middle of tall brick
            if (brick == 14) return pixelFromResource(7, 8, y); // extender for top of quad brick
            if (brick == 15) return pixelFromResource(9, 8, y); // extender for bottom of quad brick
            return 0;
        }

        private int pixelFromResource(int brick, int x, int y)
        {
            y = y * 2;
            x = x * 2 + 33 * brick;
            Color color = WindowsFormsApplication1.Properties.Resources.bricks.GetPixel(x, y);
            if (color.R == 168) return 2;
            if (color.R >= 248) return 3;
            if (color.R == 80) return 1;
            return 0;
        }

        Bitmap[] colorBricks = new Bitmap[256];

        private void drawColorBrick(int x, int y, Graphics g, int v)
        {
            if (colorBricks[v] == null)
            {
                colorBricks[v] = new Bitmap(32, 16);
                int color = v >> 4;
                int brick = v & 15;
                Color[] colors = new Color[] { Color.Black, getColor(color, 1), getColor(color, 2), getColor(color, 3) };
                for (int xx = 0; xx < 16; xx++)
                {
                    for (int yy = 0; yy < 8; yy++)
                    {
                        Color c = colors[getPixel(brick, xx, yy)];
                        colorBricks[v].SetPixel(xx * 2, yy * 2, c);
                        colorBricks[v].SetPixel(xx * 2 + 1, yy * 2, c);
                        colorBricks[v].SetPixel(xx * 2, yy * 2 + 1, c);
                        colorBricks[v].SetPixel(xx * 2 + 1, yy * 2 + 1, c);
                    }
                }
            }
            g.DrawImage(colorBricks[v], x, y);
        }

        private void panelSelectedBrick_Paint(object sender, PaintEventArgs e)
        {
            if (levels.levelType == LevelType.TI84CE)
            {
                for (int x = 0; x <= 15; x++)
                {
                    Rectangle outer = new Rectangle(x * 36, 0, 36, 20);
                    e.Graphics.FillRectangle(new SolidBrush(x == selectedBrick ? Color.Red : Color.White), outer);
                    drawColorBrick(x * 36 + 2, 2, e.Graphics, (selectedColor << 4) + x);
                }
                return;
            }

            int startBrick = 16;
            if (levels.levelType == LevelType.TI89) startBrick = 0;
            for (int x = 0; x <= 15; x++)
            {
                Rectangle outer = new Rectangle(x * 36, 0, 36, 20);
                e.Graphics.FillRectangle(new SolidBrush(x == selectedBrick ? Color.Red : Color.White), outer);
                Rectangle inner = new Rectangle(x * 36 + 2, 2, 32, 16);
                int brick = x + startBrick;
                Rectangle srcRect = new Rectangle(brick * 33, 0, 32, 16);
                e.Graphics.DrawImage(WindowsFormsApplication1.Properties.Resources.bricks, inner, srcRect, GraphicsUnit.Pixel);
            }
        }

        private void panelSelectedBrick_Click(object sender, EventArgs e)
        {
            MouseEventArgs me = e as MouseEventArgs;
            if (me != null)
            {
                int x = me.X / 36;
                if (x >= 0 && x <= levels.GetMaxBrick())
                {
                    selectedBrick = x;
                    panelSelectedBrick.Refresh();
                }
            }
        }

        private void panelLevel_Click(object sender, EventArgs e)
        {
            MouseEventArgs me = e as MouseEventArgs;
            if (me != null)
            {
                int brick = selectedBrick;
                if (levels.levelType == LevelType.TI84CE) brick += selectedColor << 4;
                if (me.Button == MouseButtons.Right) brick = 0;

                int x = me.X / 32;
                int y = me.Y / 16;
                int adjustedX = x - (20 - levels.GetWidth()) / 2;
                if (adjustedX >= 0 && adjustedX < levels.GetWidth() && y >= 0 && y < levels.GetHeight())
                {
                    levels.levels[currentLevel][adjustedX, y] = brick;
                    modifiedSinceLastSave = true;
                    Graphics g = panelLevel.CreateGraphics();
                    drawBrick(g, x, y);
                    g.Dispose();
                }
            }
        }

        private void button2_Click(object sender, EventArgs e)
        {
            if (currentLevel > 0)
            {
                currentLevel--;
                refreshCount();
            }
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (modifiedSinceLastSave && levels.IsNonTrivial() && !ShouldContinue()) return;
            OpenFileDialog dialog = new OpenFileDialog();
            dialog.Title = "Select level file to open";
            dialog.Filter = "Calculator files (*.82P;*.83P;*.8XP;*.8XV;*.85S;*.86P;*.89Y;*.9XY;*.I)|*.82P;*.83P;*.8XP;*.8XV;*.85S;*.86P;*.89Y;*.9XY;*.I";
            if (DialogResult.OK == dialog.ShowDialog())
            {
                try
                {
                    LevelData newLevels;
                    Stream infile = dialog.OpenFile();
                    if (dialog.FileName.ToUpper().EndsWith("I"))
                    {
                        newLevels = LevelData.LoadFromFile(new StreamReader(infile));
                    }
                    else
                    {
                        newLevels = LevelData.LoadFromFile(infile);
                    }
                    infile.Close();
                    levels = newLevels;
                    modifiedSinceLastSave = false;
                    currentLevel = 0;
                    fileName = dialog.FileName;
                    refreshCount();
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message, "Error loading level file", MessageBoxButtons.OK);
                }
            }
        }

        private void save(string fileName)
        {
            try
            {
                levels.Save(fileName);
                modifiedSinceLastSave = false;
                this.fileName = fileName;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Error saving level file", MessageBoxButtons.OK);
            }
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (fileName == " ")
            {
                saveAsToolStripMenuItem_Click(sender, e);
            }
            else
            {
                save(fileName);
            }
        }

        private void saveAsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SaveFileDialog dialog = new SaveFileDialog();
            dialog.Title = "Select file to save level as";
            dialog.Filter = levels.GetFilter();
            if (DialogResult.OK == dialog.ShowDialog())
            {
                save(dialog.FileName);
            }
        }

        private void button1_Click_1(object sender, EventArgs e)
        {
            if (levels.IsNonTrivial(currentLevel) && !ConfirmClear()) return;
            for (int x = 0; x < levels.GetWidth(); x++)
                for (int y = 0; y < levels.GetHeight(); y++)
                    levels.levels[currentLevel][x, y] = 0;
            panelLevel.Refresh();
        }

        private void button3_Click(object sender, EventArgs e)
        {
            if (levels.levels.Count > levels.GetLevelMin())
            {
                if (levels.IsNonTrivial(currentLevel) && !ConfirmClear()) return;
                levels.levels.RemoveAt(currentLevel);
                if (currentLevel == levels.levels.Count) currentLevel--;
                refreshCount();
            }
        }

        private void button2_Click_1(object sender, EventArgs e)
        {
            levels.InsertLevel(currentLevel);
            refreshCount();
        }

        private void labelCount_Click(object sender, EventArgs e)
        {

        }

        private void button1_Click_2(object sender, EventArgs e)
        {
            levels.InsertLevel(currentLevel);
            for (int x = 0; x < levels.GetWidth(); x++)
                for (int y = 0; y < levels.GetHeight(); y++)
                    levels.levels[currentLevel][x, y] = levels.levels[currentLevel + 1][x, y];
            currentLevel++;
            refreshCount();
        }

        private void orzunoidLevelsTI86ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            clearData(LevelType.TI86);
        }

        private void quitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (modifiedSinceLastSave && levels.IsNonTrivial() && !ShouldContinue()) return;
            Application.Exit();
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (modifiedSinceLastSave && levels.IsNonTrivial() && !ShouldContinue()) e.Cancel = true;
        }

        private void cMonsterLevelsCMonsterTI84PlusColorEditionsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            clearData(LevelType.TI84CE);
        }

        private int selectedColor = 0;

        private void panelColor_Click(object sender, EventArgs e)
        {
            MouseEventArgs me = e as MouseEventArgs;
            if (me != null)
            {
                int x = me.X / 36;
                selectedColor = x;
                panelSelectedBrick.Refresh();
            }
        }

        private void panelColor_Paint(object sender, PaintEventArgs e)
        {
            for (int x = 0; x <= 15; x++)
            {
                Rectangle inner = new Rectangle(x * 36 + 2, 2, 32, 16);
                e.Graphics.FillRectangle(new SolidBrush(getColor(x, 2)), inner);
            }
        }

        private string binary(int v, int chars)
        {
            if (chars == 0) return "";
            return binary(v >> 1, chars - 1) + (v & 1);
        }

        private void generateColorPaletteToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SaveFileDialog dialog = new SaveFileDialog();
            dialog.Title = "Select file to save color palette";
            dialog.Filter = "include files|*.I";
            if (DialogResult.OK == dialog.ShowDialog())
            {
                StreamWriter writer = new StreamWriter(dialog.FileName);
                for (int palette = 0; palette < 16; palette++)
                {
                    for (int c = 0; c < 4; c++)
                    {
                        Color color = getColor(palette, c);
                        string name = binary(color.R >> 3, 5) + binary(color.G >> 2, 6) + binary(color.B >> 3, 5);
                        writer.WriteLine("        .db     %" + name.Substring(0, 8) + ",%" + name.Substring(8, 8));
                    }
                }
                writer.Close();
            }
        }

        private void generateBrickBitmapsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SaveFileDialog dialog = new SaveFileDialog();
            dialog.Title = "Select file to save brick bitmap";
            dialog.Filter = "include files|*.I";
            if (DialogResult.OK == dialog.ShowDialog())
            {
                StreamWriter writer = new StreamWriter(dialog.FileName);
                for (int brick = 0; brick <= levels.GetMaxBrick(); brick++)
                {
                    for (int y = 0; y < 8; y++)
                    {
                        writer.Write("        .db     ");
                        for (int x = 0; x < 16; x++)
                        {
                            int color = getPixel(brick, x, y);
                            if ((x & 3) == 0 && x != 0) writer.Write(",");
                            if ((x & 3) == 0) writer.Write("%");
                            writer.Write("00011011".Substring(color * 2, 2));
                        }
                        writer.WriteLine();
                    }
                }
                writer.Close();
            }
        }

        private void shiftLeftToolStripMenuItem_Click(object sender, EventArgs e)
        {
            int blankSides = (20 - levels.GetWidth()) / 2;
            for (int x = 0; x < levels.GetWidth(); x++)
            {
                for (int y = 0; y < levels.GetHeight(); y++)
                {
                    levels.levels[currentLevel][x, y] = x == levels.GetWidth() - 1 ? 0 : levels.levels[currentLevel][x + 1, y];
                }
            }
            panelLevel.Refresh();
        }

        private void cMonsterInternalLevelsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SaveFileDialog dialog = new SaveFileDialog();
            dialog.Title = "Select file to save internal levels";
            dialog.Filter = "include files|*.I";
            if (DialogResult.OK == dialog.ShowDialog())
            {
            }
        }

        private void shiftDownToolStripMenuItem_Click(object sender, EventArgs e)
        {
            int blankSides = (20 - levels.GetWidth()) / 2;
            for (int x = 0; x < levels.GetWidth(); x++)
            {
                for (int y = levels.GetHeight() - 1; y >= 0; y--)
                {
                    levels.levels[currentLevel][x, y] = y == 0 ? 0 : levels.levels[currentLevel][x, y - 1];
                }
            }
            panelLevel.Refresh();
        }
    }
}
