﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace WindowsFormsApplication1
{
    enum LevelType
    {
        TI82,
        TI85,
        TI86,
        TI89,
        TI84CE
    };

    class LevelData
    {
        public LevelType levelType;
        public List<int[,]> levels;

        public string Description()
        {
            switch (levelType)
            {
                case LevelType.TI82:
                    return "Now editing for Orzunoid (TI-82 / TI-83 / TI-83 Plus / TI-84 Plus)";
                case LevelType.TI85:
                    return "Now editing for Orzunoid (TI-85)";
                case LevelType.TI86:
                    return "Now editing for Orzunoid (TI-86)";
                case LevelType.TI84CE:
                    return "Now editing for CMonster (TI-84 Plus color editions)";
            }
            return "Now editing for Monster (TI-89 / TI-92 / TI-92 Plus)";
        }

        public int GetLevelMin()
        {
            switch (levelType)
            {
                case LevelType.TI82:
                    return 5;
                case LevelType.TI86:
                    return 3;
            }
            return 1;
        }

        public int GetLevelMax()
        {
            switch (levelType)
            {
                case LevelType.TI82:
                    return 60;
                case LevelType.TI85:
                    return 60;
                case LevelType.TI86:
                    return 3;
                case LevelType.TI84CE:
                    return 204;
            }
            return 999;
        }

        public int GetWidth()
        {
            switch (levelType)
            {
                case LevelType.TI82:
                    return 12;
                case LevelType.TI89:
                    return 18;
                case LevelType.TI84CE:
                    return 20;
            }
            return 16;
        }

        public int GetHeight()
        {
            switch (levelType)
            {
                case LevelType.TI89:
                    return 12;
                case LevelType.TI84CE:
                    return 16;
            }
            return 8;
        }

        public int GetMaxBrick()
        {
            switch (levelType)
            {
                case LevelType.TI89:
                    return 12;
                case LevelType.TI84CE:
                    return 16;
            }
            return 3;
        }

        public LevelData(LevelType t)
        {
            levelType = t;
            levels = new List<int[,]>();
            for (int i = 0; i < GetLevelMin(); i++)
                AddLevel();
        }

        public bool AddLevel()
        {
            return InsertLevel(levels.Count);
        }

        public bool InsertLevel(int pos)
        {
            if (levels.Count == GetLevelMax()) return false;
            levels.Insert(pos, new int[GetWidth(), GetHeight()]);
            return true;
        }

        public bool IsNonTrivial(int level)
        {
            int count = 0;
            for (int x = 0; x < GetWidth(); x++)
                for (int y = 0; y < GetHeight(); y++)
                    if (levels[level][x, y] != 0) count++;
            return (count > 1);
        }

        public bool IsNonTrivial()
        {
            for (int level = 0; level < levels.Count; level++)
                if (IsNonTrivial(level)) return true;
            return false;
        }

        private void ReadLevelCompact(byte[] data, int level)
        {
            int offset = 0;
            for (int y = 0; y < GetHeight(); y++)
            {
                for (int x = 0; x < GetWidth(); x += 4)
                {
                    byte blocks = data[offset++];
                    for (int xx = x; xx < x + 4; xx++)
                    {
                        levels[level][xx, y] = (blocks >> 6) & 3;
                        blocks <<= 2;
                    }
                }
            }
        }

        public static LevelData LoadFromFile(StreamReader infile)
        {
            string initialLine = infile.ReadLine();
            if (initialLine.StartsWith(";"))
            {
                // importing old graphics studio data
                LevelData levels = new LevelData(LevelType.TI84CE);
                for (int levelNum = 0; levelNum < 25; levelNum++)
                {
                    if (levelNum != 0) levels.AddLevel();
                    {
                        int offset = 0;
                        while (offset < 384)
                        {
                            string line = infile.ReadLine();
                            for (int i = 0; i < line.Length; i++)
                            {
                                if (line[i] == '$')
                                {
                                    for (int ii = i + 1; ii <= i + 2; ii++)
                                    {
                                        int nibble = "0123456789ABCDEF".IndexOf(line[ii]);
                                        for (int bit = 3; bit >= 0; bit--)
                                        {
                                            int val = nibble & (1 << bit);
                                            if (val != 0)
                                            {
                                                int x = offset % 24;
                                                int y = offset / 24;
                                                int color = ((x ^ y) & 3) + 1;
                                                if (x < 20)
                                                {
                                                    levels.levels[levelNum][x, y] = color;
                                                }
                                            }
                                            offset++;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                return levels;
            }
            if (initialLine.StartsWith("NUM_LEVELS"))
            {
                // old 4-bit include file
                LevelData levels = new LevelData(LevelType.TI84CE);
                int levelCount = Int32.Parse(initialLine.Substring(initialLine.IndexOf('=') + 1));
                for (int levelNum = 0; levelNum < levelCount; levelNum++)
                {
                    if (levelNum != 0) levels.AddLevel();
                    {
                        int offset = 0;
                        while (offset < 320)
                        {
                            string line = infile.ReadLine();
                            for (int i = 0; i < line.Length; i++)
                            {
                                if (line[i] == '$')
                                {
                                    for (int ii = i + 1; ii <= i + 2; ii++)
                                    {
                                        int[] conv = { 0, 1, 0x31, 0x91, 0xc1, 2, 0x32, 0x92, 0xc2, 3 };
                                        int nibble = "0123456789ABCDEF".IndexOf(line[ii]);
                                        int x = offset % 20;
                                        int y = offset / 20;
                                        levels.levels[levelNum][x, y] = conv[nibble];
                                        offset++;
                                    }
                                }
                            }
                        }
                    }
                }
                return levels;
            }
            if (initialLine.StartsWith("#define"))
            {
                // 8-bit include file
                LevelData levels = new LevelData(LevelType.TI84CE);
                int levelCount = Int32.Parse(initialLine.Substring(initialLine.IndexOf("LEVELS") + 6));
                for (int levelNum = 0; levelNum < levelCount; levelNum++)
                {
                    if (levelNum != 0) levels.AddLevel();
                    {
                        int offset = 0;
                        while (offset < 320)
                        {
                            string line = infile.ReadLine();
                            if (line.StartsWith(";")) continue;
                            for (int i = 0; i < line.Length; i++)
                            {
                                if (Char.IsDigit(line[i]))
                                {
                                    int next = line.IndexOf(',', i);
                                    if (next == -1) next = line.Length;
                                    int x = offset % 20;
                                    int y = offset / 20;
                                    levels.levels[levelNum][x, y] = Int32.Parse(line.Substring(i, next - i));
                                    offset++;
                                    i = next;
                                }
                            }
                        }
                    }
                }
                return levels;
            }
            throw new Exception("Unrecognized file type");
        }

        public static LevelData LoadFromFile(Stream infile)
        {
            Encoding enc = new UTF8Encoding(true, true);
            LevelData levels;
            byte[] signature = new byte[8];
            infile.Read(signature, 0, 8);
            string strSig = enc.GetString(signature);
            int levelCount;

            switch (strSig)
            {
                case "**TI89**":
                case "**TI92P*":
                    infile.Seek(0x58, SeekOrigin.Begin);
                    levelCount = infile.ReadByte();
                    levelCount *= 256;
                    levelCount += infile.ReadByte();
                    if (infile.ReadByte() != 0x6a) throw new Exception("Not a Monster level file");
                    if (infile.ReadByte() != 0x4f) throw new Exception("Not a Monster level file");
                    infile.Seek(0x76, SeekOrigin.Begin);
                    levels = new LevelData(LevelType.TI89);
                    for (int level = 0; level < levelCount; level++)
                    {
                        if (level != 0) levels.AddLevel();
                        for (int y = 0; y < 12; y++)
                            for (int x = 0; x < 18; x++)
                                levels.levels[level][x, y] = infile.ReadByte();
                    }
                    break;

                case "**TI82**":
                case "**TI83**":
                case "**TI83F*":
                    infile.Seek(strSig == "**TI83F*" ? 0x4a : 0x48, SeekOrigin.Begin);
                    switch (infile.ReadByte())
                    {
                        case 5:
                            if (infile.ReadByte() != 0) throw new Exception("Not a recognized level file");
                            levels = new LevelData(LevelType.TI82);
                            for (int level = 0; level < 5; level++)
                            {
                                byte[] data = new byte[24];
                                infile.Read(data, 0, 24);
                                levels.ReadLevelCompact(data, level);
                            }
                            for (int level = 5; level < levels.GetLevelMax(); level++)
                            {
                                byte[] data = new byte[24];
                                try
                                {
                                    if (24 != infile.Read(data, 0, 24)) break;
                                }
                                catch
                                {
                                    break;
                                }
                                levels.AddLevel();
                                levels.ReadLevelCompact(data, level);
                            }
                            break;
                        case 0x43:
                            levels = new LevelData(LevelType.TI84CE);
                            byte[] sig = new byte[7];
                            infile.Read(sig, 0, 7);
                            if (Encoding.UTF8.GetString(sig) != "MonLvls") throw new Exception("Not a recognized level file");
                            for (int level = 0; infile.Position + 320 <= infile.Length; level++)                                
                            {
                                if (level != 0) levels.AddLevel();
                                for (int y = 0; y < levels.GetHeight(); y++)
                                {
                                    for (int x = 0; x < levels.GetWidth(); x++)
                                    {
                                        levels.levels[level][x, y] = infile.ReadByte();
                                    }
                                }
                            }
                            if (levels.levels.Count > levels.GetLevelMax()) throw new Exception("Level file is too long");
                            break;
                        default:
                            throw new Exception("Not a recognized level file");
                    }
                    break;

                case "**TI86**":
                    infile.Seek(0x54, SeekOrigin.Begin);
                    if (infile.ReadByte() != 0xd0) throw new Exception("Not a recognized Orzunoid level file");
                    if (infile.ReadByte() != 0x4f) throw new Exception("Not a recognized Orzunoid level file");
                    infile.Seek(0x65, SeekOrigin.Begin);
                    while (infile.ReadByte() != 0) ;
                    levels = new LevelData(LevelType.TI86);
                    for (int level = 0; level < 3; level++)
                    {
                        byte[] data = new byte[32];
                        infile.Read(data, 0, 32);
                        levels.ReadLevelCompact(data, level);
                    }
                    break;

                case "**TI85**":
                    byte[] rawData = new byte[30000];
                    infile.Read(rawData, 0, 30000);
                    int i = 0;
                    // seek signature word
                    while (i++ < 29997)
                        if (rawData[i] == 0xCA && rawData[i + 1] == 0xCD)
                            break;
                    // seek level count (looking for instruction to store it)
                    while (i++ < 29997)
                        if (rawData[i] == 0x32 && rawData[i + 1] == 0x90 && rawData[i + 2] == 0x8A)
                            break;
                    levelCount = rawData[i - 1];
                    // seek level data (looking for end of extraction code);
                    while (i++ < 29997)
                        if (rawData[i] == 0x10 && rawData[i + 1] == 0xED && rawData[i + 2] == 0xC9)
                            break;
                    if (i == 29997) throw new Exception("Not a recognized level file format");
                    infile.Seek(i + 11, SeekOrigin.Begin);
                    levels = new LevelData(LevelType.TI85);
                    for (int level = 0; level < levelCount; level++)
                    {
                        byte[] data = new byte[32];
                        infile.Read(data, 0, 32);
                        if (level != 0) levels.AddLevel();
                        levels.ReadLevelCompact(data, level);
                    }
                    break;

                default:
                    throw new Exception("Unrecognized file type");
            }                    
            return levels;
        }

        public string GetFilter()
        {
            switch (levelType)
            {
                case LevelType.TI82:
                    return "TI-82 files (*.82P)|*.82P|TI-83 files (*.83P)|*.83P|TI-83 Plus files (*.8XP)|*.8XP";
                case LevelType.TI85:
                    return "TI-85 files (*.85S)|*.85S";
                case LevelType.TI86:
                    return "TI-86 files (*.86P)|*.86P";
                case LevelType.TI84CE:
                    return "TI-84 CSE/CE files (*.8XV)|*.8XV";
                default:
                    return "TI-89 files (*.89Y)|*.89Y|TI-92 Plus files (*.9XY)|*.9XY";
            }
        }

        public void Save(string fileName)
        {
            string ext = Path.GetExtension(fileName);

            if (ext == ".I")
            {
                StreamWriter writer = new StreamWriter(fileName);
                //writer.WriteLine("NUM_LEVELS = " + levels.Count);
                writer.WriteLine("#define NUM_LEVELS " + levels.Count);
                for (int level = 0; level < levels.Count; level++)
                {
                    writer.WriteLine("; Level " + (level + 1));
                    for (int y = 0; y < 16; y++)
                    {
                        writer.Write("        .db     ");
                        for (int x = 0; x < 20; x += 1)
                        {
                            int brick = levels[level][x, y];
                            writer.Write(brick);
                            if (x == 19)
                            {
                                writer.WriteLine();
                            }
                            else
                            {
                                writer.Write(",");
                            }
                        }
                        /*for (int x = 0; x < 20; x += 2)
                        {
                            writer.Write("$");
                            int firstBrick = levels[level][x, y];
                            int secondBrick = levels[level][x + 1, y];
                            writer.Write("0123456789ABCDEF"[firstBrick]);
                            writer.Write("0123456789ABCDEF"[secondBrick]);
                            if (x == 18)
                            {
                                writer.WriteLine();
                            }
                            else
                            {
                                writer.Write(",");
                            }
                        }
                        */
                    }
                }
                writer.Close();
                return;
            }

            string varName = Path.GetFileNameWithoutExtension(fileName);
            VariableWriter v = null;
            switch (levelType)
            {
                case LevelType.TI82:
                    if (ext.ToUpper() == ".82P") v = new VariableWriter(VariableType.TI82);
                    if (ext.ToUpper() == ".83P") v = new VariableWriter(VariableType.TI83);
                    if (ext.ToUpper() == ".8XP") v = new VariableWriter(VariableType.TI83P);
                    break;
                case LevelType.TI84CE:
                    if (ext.ToUpper() == ".8XV") v = new VariableWriter(VariableType.TI84CE);
                    break;
                case LevelType.TI85:
                    if (ext.ToUpper() == ".85S") v = new VariableWriter(VariableType.TI85);
                    break;
                case LevelType.TI86:
                    if (ext.ToUpper() == ".86P") v = new VariableWriter(VariableType.TI86);
                    break;
                case LevelType.TI89:
                    if (ext.ToUpper() == ".89Y") v = new VariableWriter(VariableType.TI89);
                    if (ext.ToUpper() == ".9XY") v = new VariableWriter(VariableType.TI92P);
                    break;
            }
            if (v == null)
            {
                throw new Exception("invalid file extension");
            }

            v.SetVariableName(varName);
            switch (levelType)
            {
                case LevelType.TI84CE:
                    v.variableData.AddRange(Encoding.UTF8.GetBytes("CMonLvls"));
                    for (int level = 0; level < levels.Count; level++)
                        for (int y = 0; y < GetHeight(); y++)
                            for (int x = 0; x < GetWidth(); x++)
                                v.variableData.Add((byte)levels[level][x, y]);
                    break;
                case LevelType.TI82:
                    v.variableData.Add(5);
                    v.variableData.Add(0);
                    for (int level = 0; level < levels.Count; level++)
                        WriteLevelCompact(level, v);
                    v.variableData.Add(1);
                    v.variableData.Add(0xa7);
                    break;
                case LevelType.TI85:
                     byte[] initBytes = {0x00, 0xf8, 0x17, 0x4c, 0x65, 0x76, 0x65, 0x6c,
 0x73, 0x20, 0x66, 0x6f, 0x72, 0x20, 0x4f, 0x72, 0x75, 0x7a, 0x6e, 0x6f, 0x69, 0x64, 0x20, 0x36,
 0x2b, 0x00, 0x11, 0x6b, 0x8a, 0x01, 0x0e, 0x00, 0x21, 0x3d, 0x00, 0xed, 0xb0, 0x21, 0x34, 0x00,
 0xcd, 0xaa, 0x8c, 0x21, 0x00, 0x00, 0x22, 0x6b, 0x8a, 0x3e, 0x03, 0x32, 0x40, 0x8c, 0xc9, 0x4f,
 0x72, 0x7a, 0x75, 0x6e, 0x6f, 0x69, 0x64, 0x00, 0xca, 0xcd, 0x4b, 0x00};
                    for (int i = 0; i < initBytes.Length; i++) 
                    v.variableData.Add(initBytes[i]);
                    for (int i = 0; i < 8; i++)
                    {
                        if (varName.Length > i)
                        {
                            v.variableData.Add(Encoding.UTF8.GetBytes(varName)[i]);
                        }
                        else
                        {
                            v.variableData.Add(0);
                        }
                    }
                    byte[] midBytes = {0x00, 0x00, 0x21, 0x57, 0x00, 0x22, 0x85, 0x8a, 0x3e, 0x08, 0x32, 0x90,
0x8a, 0xc9, 0x21, 0x5b, 0x00, 0x47, 0x04, 0x11, 0x20, 0x00, 0x19, 0x10, 0xfd, 0x06, 0x20, 0x11,
0x00, 0x81, 0xc5, 0x06, 0x04, 0x4e, 0xcb, 0x01, 0xcb, 0x01, 0x79, 0xe6, 0x03, 0x12, 0x13, 0x10,
0xf5, 0x23, 0xc1, 0x10, 0xed, 0xc9};
                    midBytes[9] = (byte) levels.Count;
                    for (int i = 0; i < midBytes.Length; i++)
                        v.variableData.Add(midBytes[i]);
                    for (int level = 0; level < levels.Count; level++)
                        WriteLevelCompact(level, v);
                    byte[] relocBytes = { 0x1e, 05, 0x1c, 0x0d, 0x0c, 0xfb, 0xff, 05, 00 };
                    for (int i = 0; i < relocBytes.Length; i++)
                        v.variableData.Add(relocBytes[i]);
                    break;
                case LevelType.TI86:
                    byte[] initBytes86 = {0x8e, 0x28, 0x00, 0xc3, 0x50, 0xd7, 0x00,
0x00, 0x62, 0xd7, 0x21, 0xd0, 0x4f, 0x22, 0x12, 0xc0, 0x11, 0x14, 0xc0, 0x21, 0x7c, 0xd7, 0x01,
0x60, 0x00, 0xed, 0xb0, 0xc9, 0x45, 0x78, 0x74, 0x72, 0x61, 0x20, 0x6c, 0x65, 0x76, 0x65, 0x6c,
0x73, 0x20, 0x66, 0x6f, 0x72, 0x20, 0x4f, 0x72, 0x7a, 0x75, 0x6e, 0x6f, 0x69, 0x64, 0x00};
                    for (int i = 0; i < initBytes86.Length; i++)
                        v.variableData.Add(initBytes86[i]);
                    for (int level = 0; level < levels.Count; level++)
                        WriteLevelCompact(level, v);
                    break;
                case LevelType.TI89:
                    v.variableData.Add((byte) (levels.Count >> 8)); // level count, stored as big endian
                    v.variableData.Add((byte) (levels.Count % 0x100));
                    v.variableData.Add(0x6A); // signature for Monster levels
                    v.variableData.Add(0x4F);
                    for (int i = 0; i < 8; i++)
                    {
                        if (varName.Length > i)
                        {
                            v.variableData.Add(Encoding.UTF8.GetBytes(varName)[i]);
                        }
                        else
                        {
                            v.variableData.Add(0);
                        }
                    }
                    v.variableData.Add(0x0C); // additional level identifier
                    v.variableData.Add(0xED);
                    for (int i = 0; i < 16; i++) v.variableData.Add(0); // 16 bytes of reserved data

                    // unlike other calculators, 89/92+ level data is not compacted
                    for (int level = 0; level < levels.Count; level++)
                        for (int y = 0; y < GetHeight(); y++)
                            for (int x = 0; x < GetWidth(); x++)
                                v.variableData.Add((byte)levels[level][x, y]);

                    v.variableData.Add(0);      // custom data type (mlv)
                    v.variableData.Add((byte) 'm');
                    v.variableData.Add((byte) 'l');
                    v.variableData.Add((byte) 'v');
                    v.variableData.Add(0);
                    v.variableData.Add(0xF8);
                    break;
            }
            v.WriteToFile(fileName);
        }

        private void WriteLevelCompact(int level, VariableWriter v)
        {
            for (int y = 0; y < GetHeight(); y++)
            {
                for (int x = 0; x < GetWidth(); x += 4)
                {
                    byte blocks = 0;
                    for (int xx = x; xx < x + 4; xx++)
                    {
                        blocks <<= 2;
                        blocks |= (byte)levels[level][xx, y];
                    }
                    v.variableData.Add(blocks);
                }
            }
        }
    }
}
