﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace WindowsFormsApplication1
{
    enum VariableType
    {
        TI82,
        TI83,
        TI83P,
        TI84CE,
        TI85,
        TI86,
        TI89,
        TI92P
    };

    class VariableWriter
    {
        private List<byte> variableHeader;
        public List<byte> variableData;
        private VariableType type;
        private string variableName;

        public VariableWriter(VariableType t)
        {
            type = t;
            variableData = new List<byte>();
            variableHeader = new List<byte>();
        }

        public void SetVariableName(string name)
        {
            variableName = "";
            if (type == VariableType.TI82 || type == VariableType.TI83 || type == VariableType.TI83P) name = name.ToUpper();
            if (type == VariableType.TI89 || type == VariableType.TI92P) name = name.ToLower();
            for (int x = 0; x < name.Length; x++)
            {
                if (variableName.Length == 8) break;
                if (name[x] >= 'a' && name[x] <= 'z')
                    variableName += name[x];
                if (name[x] >= 'A' && name[x] <= 'Z')
                    variableName += name[x];
                if (name[x] >= '0' && name[x] <= '9')
                    variableName += name[x];
            }
            if (variableName.Length == 0) throw new Exception("Invalid variable name");
        }

        public void WriteToFile(string fileName)
        {
            Stream outFile = new FileStream(fileName, FileMode.Create);
            try
            {
                Encoding enc = Encoding.UTF8;
                string header = "";

                // Write out identification in header
                switch (type)
                {
                    case VariableType.TI82:
                        header = "**TI82**";
                        break;
                    case VariableType.TI83:
                        header = "**TI82**";
                        break;
                    case VariableType.TI83P:
                    case VariableType.TI84CE:
                        header = "**TI83F*";
                        break;
                    case VariableType.TI85:
                        header = "**TI85**";
                        break;
                    case VariableType.TI86:
                        header = "**TI86**";
                        break;
                    case VariableType.TI89:
                        header = "**TI89**";
                        break;
                    case VariableType.TI92P:
                        header = "**TI92P*";
                        break;
                }
                outFile.Write(enc.GetBytes(header), 0, 8);

                // Further header data and descriptive descriptive comment
                string comment = "";
                if (type == VariableType.TI89 || type == VariableType.TI92P)
                {
                    outFile.WriteByte(1);
                    outFile.WriteByte(0);
                    outFile.Write(enc.GetBytes("main"), 0, 4); // on 89/92+ folder precedes the comment
                    for (int i = 0; i < 4; i++) outFile.WriteByte(0);
                    comment = "Monster level file                      ";
                }
                else if (type == VariableType.TI84CE)
                {
                    outFile.WriteByte(0x1a);
                    outFile.WriteByte(0xa);
                    outFile.WriteByte(0xa);
                    comment = "CMonster level file                       ";
                }
                else
                {
                    outFile.WriteByte(0x1a);
                    outFile.WriteByte(0xa);
                    outFile.WriteByte(0);
                    comment = "Orzunoid level file                       ";
                }

                outFile.Write(enc.GetBytes(comment), 0, comment.Length);

                if (type == VariableType.TI89 || type == VariableType.TI92P)
                {
                    outFile.WriteByte(1); // number of variables
                    outFile.WriteByte(0);
                    outFile.WriteByte(0x52);
                    outFile.WriteByte(0);
                    outFile.WriteByte(0);
                    outFile.WriteByte(0);
                }
                else
                {
                    // variable header size, will be filled in later
                    variableHeader.Add(0);
                    variableHeader.Add(0);

                    // data size (including length word at beginning of data)
                    variableHeader.Add((byte)((variableData.Count + 2) & 0xFF));
                    variableHeader.Add((byte)((variableData.Count + 2) >> 8));

                    // variable type
                    switch (type)
                    {
                        case VariableType.TI82:
                        case VariableType.TI83:
                        case VariableType.TI83P:
                            variableHeader.Add(6);
                            break;
                        case VariableType.TI85:
                            variableHeader.Add(12);
                            break;
                        case VariableType.TI86:
                            variableHeader.Add(18);
                            break;
                        case VariableType.TI84CE:
                            variableHeader.Add(21);
                            break;
                    }
                }

                // variable name
                if (type == VariableType.TI85 || type == VariableType.TI86)
                {
                    variableHeader.Add((byte)variableName.Length);
                }
                byte[] nameBytes = enc.GetBytes(variableName);
                for (int i = 0; i < 8; i++)
                {
                    if (i < variableName.Length)
                    {
                        variableHeader.Add(nameBytes[i]);
                    }
                    else if (type != VariableType.TI85)
                    {
                        variableHeader.Add(0);
                    }
                }

                if (type == VariableType.TI89 || type == VariableType.TI92P)
                {
                    // on these calcs variable type is after the name
                    variableHeader.Add(0x1C); // indicates custom type
                    variableHeader.Add(0);
                    variableHeader.Add(0);
                    variableHeader.Add(0);
                    variableHeader.Add(0xC6); // file length will go here
                    variableHeader.Add(9);
                    variableHeader.Add(0);
                    variableHeader.Add(0);

                    variableHeader.Add(0xa5);
                    variableHeader.Add(0x5a);
                    variableHeader.Add(0);
                    variableHeader.Add(0);
                    variableHeader.Add(0);
                    variableHeader.Add(0);
                }

                // extra bytes for 83+ / 84+ only
                if (type == VariableType.TI83P || type == VariableType.TI84CE)
                {
                    variableHeader.Add(0);
                    variableHeader.Add(0);
                }

                if (type == VariableType.TI89 || type == VariableType.TI92P)
                {
                    // data size (not including length word at beginning of data) - this is big endian on 89/92+
                    variableHeader.Add((byte)(variableData.Count >> 8)); 
                    variableHeader.Add((byte)(variableData.Count & 0xFF));
                }
                else
                {
                    // data size (including length word at beginning of data)
                    variableHeader.Add((byte)((variableData.Count + 2) & 0xFF));
                    variableHeader.Add((byte)((variableData.Count + 2) >> 8));

                    // data size (not including length word at beginning of data)
                    variableHeader.Add((byte)(variableData.Count & 0xFF));
                    variableHeader.Add((byte)(variableData.Count >> 8));
                }

                if (type == VariableType.TI89 || type == VariableType.TI92P)
                {
                    // insert full file length in header - this is little endian
                    int filelen = 90 + variableData.Count;
                    variableHeader[13] = (byte)(filelen & 0xFF);
                    variableHeader[12] = (byte)(filelen >> 8);
                }
                else
                {
                    // fill in header size at beginning of variable header
                    variableHeader[0] = (byte)(variableHeader.Count - ((type == VariableType.TI85 || type == VariableType.TI86 || type == VariableType.TI83P) ? 6 : 4));

                    // write out combined size value
                    int combinedCount = variableHeader.Count + variableData.Count;
                    outFile.WriteByte((byte)(combinedCount & 0xFF));
                    outFile.WriteByte((byte)(combinedCount >> 8));
                }

                // write out the data and checksum
                int checksum = 0;
                for (int x = 0; x < variableHeader.Count; x++)
                {
                    checksum += variableHeader[x];
                    outFile.WriteByte(variableHeader[x]);
                }
                for (int x = 0; x < variableData.Count; x++)
                {
                    checksum += variableData[x];
                    outFile.WriteByte(variableData[x]);
                }
                outFile.WriteByte((byte)(checksum & 0xFF));
                outFile.WriteByte((byte)(checksum >> 8));
            }
            finally
            {
                outFile.Close();
            }
        }
    }
}
